
/* symtab.c,v 1.1.1.1 1996/03/03 17:39:51 arsharma Exp */

/*
 *  Purpose: symbol table  management
 *  Note: A symbol table is required because, it is not possible to
 *	  tell by just looking at an ID, whether it is a memory
 *	  reference or not. Assembler constants are not memory
 *	  references, whereas variables corresponding to memory
 *	  locations are. This is required only for DOS assemblers like
 *	  MASM, but not for Unix assemblers.
 */

#include <stdio.h>
#include <stdlib.h>
#include "symtab.h"

static void expandwords ();
static void expandbuf ();

void *allocate (unsigned long n);

extern int yyerror ();
extern int yylineno;

/*
 *  treat wptr as a hash table:	      all entries are initially NULL entries
 *  that are in use are non-NULL  on collisions, search forward from
 *  the desired array entry to an empty entry.
 */
static Symbol *wptr;
static unsigned int hashsize;

/*
 *  the	      wptr array of strings is of flexible size.  its size is always
 *  one less than some power of 2.  it expands as necessary to
 *  accommodate the expanding needs of the application.
 */

#define InitHashSize	512

/*
 * we maintain permanent copies	      of strings in buf (and additional
 *  heap-allocated segments if necessary).
 */

#define BufSize	 2048
static char buf[BufSize], *nxtfree = buf, *eobuf = buf + BufSize;

/* Allocate the hash table and initialize it */
void
initwords ()
{
    int i;

    hashsize = InitHashSize - 1;
    wptr = (Symbol *) allocate (hashsize * sizeof (struct symbol));
    for (i = 0; i < hashsize; i++)
	wptr[i] = NULL;
}

/* A simple hashing function */
unsigned int
hash (register unsigned char *s)
{
    register unsigned int i = 0;

    while (*s)
	i = i + *s++;
    return i;
}

/*
 * We are out of space,	      so allocate a table double the size of the
 * original table and shift all the data into the new one
 */
static void
expandwords ()
{
    int newsize = (2 * (hashsize + 1)) - 1;
    Symbol *newtable = (Symbol *) allocate (newsize * sizeof (struct
							      symbol));
    int i;

    if (!newtable)
	yyerror ("out of heap space in string manager");

    for (i = 0; i < newsize; i++)
	newtable[i] = NULL;

    for (i = 0; i < hashsize; i++) {
	Symbol s;

	if (s = wptr[i]) {
	    int h;

	    h = hash (s->name) % newsize;

	    while (newtable[h])
		h = (h + 1) % newsize;

	    newtable[h] = s;
	}
    }

    wptr = newtable;
    hashsize = newsize;

}

/* allocate more memory for the string buffer */
static void
expandbuf ()
{
    nxtfree = allocate (BufSize);
    if (!nxtfree)
	yyerror ("out of heap space in string manager at line %d\n",
		 yylineno);

    eobuf = nxtfree + BufSize;
}

/*
 *    slookup looks up searches	      the table of existing words for the
 *	 null-terminated string s.
 *
 *	if this string is found, slookup returns a pointer to it.
 *	else
 *	  if install is non-zero, insert the string into table and return
 *	      its new location
 *	  else return NULL;
 */

Symbol
slookup (char *s, int install)
{
    int h, oldh;

    oldh = h = hash (s) % hashsize;

    while (wptr[h] && strcmp (s, wptr[h]->name)) {
	h = (h + 1) % hashsize;

	if (h == oldh) {
	    /* The hash table is full. Expand it */
	    expandwords ();

	    /* redo the lookup */
	    return slookup (s, install);
	}
    }

    /*
     * If it is	      not found and we need to install it, then check if we
     * still have space left and if not, expand the string buffer
     */
    if (!wptr[h] && install) {
	if (nxtfree + strlen (s) + 1 > eobuf)
	    expandbuf ();

	NEW0 (wptr[h]);
	wptr[h]->name = nxtfree;

	/* copy the string */
	while (*nxtfree++ = *s++);
    }

    return wptr[h];
}

/* allocate n bytes of memory */
void *
allocate (unsigned long n)
{
    return malloc (n);
}
