/*
 * Definitions of inline macros for queue manipulation and
 * memory management of fixed-sized structs using free lists.
 *
 * Copyright (C) 1993 by Jack E. Veenstra (veenstra@cs.rochester.edu)
 * 
 * This file is part of MINT, a MIPS code interpreter and event generator
 * for parallel programs.
 * 
 * MINT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 * 
 * MINT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with MINT; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __macros_h
#define __macros_h

#include "queue.h"

/* Define inline versions of the queue manipulation functions. */

/* Initialize the head node for a queue. */
#define INLINE_INIT_Q(Q) \
((Q)->next = Q, (Q)->prev = Q)

/* remove T from whatever Q it is in */
#define INLINE_REMOVE(T) \
((T)->next->prev = (T)->prev, (T)->prev->next = (T)->next)

/* enqueue T at the end of Q */
#define INLINE_ENQUEUE(Q,T) \
((T)->next = (Q), (T)->prev = (Q)->prev, (Q)->prev->next = (T), (Q)->prev = (T))

/* insert T after Q */
#define INLINE_INSERT_AFTER(Q,T) \
((T)->next = (Q)->next, (T)->prev = (Q), (Q)->next->prev = (T), (Q)->next = (T))

/* insert T before Q */
#define INLINE_INSERT_BEFORE(Q,T) \
((T)->next = (Q), (T)->prev = (Q)->prev, (Q)->prev->next = (T), (Q)->prev = (T))

/* dequeue a T from head of Q */
#define INLINE_DEQUEUE(Q,T) \
((T)=(Q)->next, (T)==(Q) ? (T)=NULL : ((Q)->next=(T)->next, (T)->next->prev=(Q)))

/* Dequeue a T from head of Q, but do not check for an empty queue.
 * If the queue was empty, then T will be set equal to Q, so the
 * program should check this after the macro is executed.
 */
#define INLINE_DEQUEUE_NOCHECK(Q,T) \
((T)=(Q)->next, (Q)->next=(T)->next, (T)->next->prev=(Q))

/* These versions of the macros are the same except that the next and prev
 * fields are specified in the macro. This allows a structure to be on
 * different queues at the same time.
 */

/* Initialize the head node for a queue. */
#define INLINE_INIT_Q_NP(Q,next,prev) \
((Q)->next = Q, (Q)->prev = Q)

/* remove T from whatever Q it is in */
#define INLINE_REMOVE_NP(T,next,prev) \
((T)->next->prev = (T)->prev, (T)->prev->next = (T)->next)

/* enqueue T at the end of Q */
#define INLINE_ENQUEUE_NP(Q,T,next,prev) \
((T)->next = (Q), (T)->prev = (Q)->prev, (Q)->prev->next = (T), (Q)->prev = (T))

/* insert T after Q */
#define INLINE_INSERT_AFTER_NP(Q,T,next,prev) \
((T)->next = (Q)->next, (T)->prev = (Q), (Q)->next->prev = (T), (Q)->next = (T))

/* insert T before Q */
#define INLINE_INSERT_BEFORE_NP(Q,T,next,prev) \
((T)->next = (Q), (T)->prev = (Q)->prev, (Q)->prev->next = (T), (Q)->prev = (T))

/* dequeue a T from head of Q */
#define INLINE_DEQUEUE_NP(Q,T,next,prev) \
((T)=(Q)->next, (T)==(Q) ? (T)=NULL : ((Q)->next=(T)->next, (T)->next->prev=(Q)))

/* number of structs to allocate in one call to malloc when out of free items */
#ifndef NITEMS
#define NITEMS 100
#endif

#if defined(DEBUG) || defined(NO_INLINING)

/* dbx chokes on multi-line macros */
#define NEW_ITEM(FREE_ITEM_PTR, ITEM_SIZE, NEW_PTR, NAME) NEW_PTR = (void *) new_item(&(FREE_ITEM_PTR), ITEM_SIZE, NAME)
#define NEW_ZITEM(FREE_ITEM_PTR, ITEM_SIZE, NEW_PTR, NAME) NEW_PTR = (void *) new_zitem(&(FREE_ITEM_PTR), ITEM_SIZE, NAME)
#define FREE_ITEM(FREE_ITEM_PTR, ITEM_PTR) free_item(&(FREE_ITEM_PTR), ITEM_PTR)

#else /* !DEBUG */

/* Define a macro for allocating a new structure. This is trickier than
 * the corresponding function definition since we do not refer to any
 * of the fields by name. (We could cast each item pointer to another type
 * of pointer with known field names, but we can also do it without
 * defining another structure.)
 */
#define NEW_ITEM(FREE_ITEM_PTR, ITEM_SIZE, NEW_PTR, NAME) \
{ \
    int _ii; \
\
    /* Reduce calls to malloc and make more efficient use of space \
     * by allocating several structs at once. \
     */ \
    if ((FREE_ITEM_PTR) == NULL) { \
        NEW_PTR = (void *) malloc((NITEMS) * (ITEM_SIZE)); \
        if ((NEW_PTR) == NULL) { \
            fprintf(stderr, "%s: out of memory\n", NAME); \
            exit(1); \
        } \
 \
        /* Link all the free structs using the first field in the struct. */ \
        /* This assumes that the structure is large enough to hold a pointer. */ \
        FREE_ITEM_PTR = (NEW_PTR); \
        for (_ii = 0; _ii < (NITEMS) - 1; _ii++) { \
            *((long **)(FREE_ITEM_PTR)) = (long *) ((char *) (NEW_PTR) + ((_ii + 1) * (ITEM_SIZE))); \
            FREE_ITEM_PTR = (void *) *(long **)(FREE_ITEM_PTR); \
        } \
        *((long **)(FREE_ITEM_PTR)) = NULL; \
        FREE_ITEM_PTR = (NEW_PTR); \
    } \
 \
    /* Remove a free item from the front of the list. */ \
    NEW_PTR = FREE_ITEM_PTR; \
    FREE_ITEM_PTR = (void *) *(long **)(FREE_ITEM_PTR); \
}

/* Define a macro for allocating a new structure, initialized to zero.
 * Zero the item only when it is about to be used, not when allocating
 * the array. This saves real memory when large items are allocated
 * and not used, and saves time since the item must be zeroed anyway
 * just before it is returned (since it may have been used and then freed).
 */
#define NEW_ZITEM(FREE_ITEM_PTR, ITEM_SIZE, NEW_PTR, NAME) \
{ \
    int _ii; \
\
    /* Reduce calls to malloc and make more efficient use of space \
     * by allocating several structs at once. \
     */ \
    if ((FREE_ITEM_PTR) == NULL) { \
        NEW_PTR = (void *) malloc((NITEMS) * (ITEM_SIZE)); \
        if ((NEW_PTR) == NULL) { \
            fprintf(stderr, "%s: out of memory\n", NAME); \
            exit(1); \
        } \
 \
        /* Link all the free structs using the first field in the struct. */ \
        /* This assumes that the structure is large enough to hold a pointer. */ \
        FREE_ITEM_PTR = (NEW_PTR); \
        for (_ii = 0; _ii < (NITEMS) - 1; _ii++) { \
            *((long **)(FREE_ITEM_PTR)) = (long *) ((char *) (NEW_PTR) + ((_ii + 1) * (ITEM_SIZE))); \
            FREE_ITEM_PTR = (void *) *(long **)(FREE_ITEM_PTR); \
        } \
        *((long **)(FREE_ITEM_PTR)) = NULL; \
        FREE_ITEM_PTR = (NEW_PTR); \
    } \
 \
    /* Remove a free item from the front of the list. */ \
    NEW_PTR = FREE_ITEM_PTR; \
    FREE_ITEM_PTR = (void *) *(long **)(FREE_ITEM_PTR); \
    memset(NEW_PTR, 0, ITEM_SIZE); \
}

/* define a macro for free-ing a new structure */
#define FREE_ITEM(FREE_ITEM_PTR, ITEM_PTR) \
{ \
    /* Put a free item on the front of the list. */ \
    *((long **) (ITEM_PTR)) = (long *) FREE_ITEM_PTR; \
    FREE_ITEM_PTR = ITEM_PTR; \
}

#endif /* !DEBUG */

#endif /* !__macros_h */
