/*
 * Header file for "page table" data structure.
 *
 * Copyright (C) 1993 by Jack E. Veenstra (veenstra@cs.rochester.edu)
 * 
 * This file is part of MINT, a MIPS code interpreter and event generator
 * for parallel programs.
 * 
 * MINT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 * 
 * MINT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with MINT; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __pt_h
#define __pt_h

#include "common.h"

/*
 * The program using the page table routines should define a structure
 * for containing page table information. The struct must include as its
 * first field an unsigned int field to store the address. Other fields
 * are optional.
 *
 * Example:
 *
 * struct pte {
 *     unsigned int addr;
 *     unsigned int accessed;
 *     unsigned int written;
 * };
 *
 *
 * #define PTE_SIZE (sizeof(struct pte))
 *
 * Dec, 1991	Jack Veenstra		veenstra@cs.rochester.edu
 *    Created.
 *
 * Oct, 1992	Jack Veenstra		veenstra@cs.rochester.edu
 *    Modified so that the generic page table routines do not need to
 * be compiled with the user's pte.h file. This allows different structs
 * to be used with the same routines. It also eliminates the dependency
 * on the user file "pte.h" and eliminates the need to recompile pt.c.
 */

/* NLEVEL1 is the number of pointers in the top level, statically allocated */
#define NLEVEL1	16384

/* NLEVEL2 is the number of pointers at the 2nd level, dynamically allocated */
#define NLEVEL2	16384

/* NLEVEL3 is the number of structs at the 3rd level, dynamically allocated */
#define NLEVEL3	16

/* The number of bits in each field must be the log of the corresponding
 * value for NLEVEL given above.
 */
#define BITS1	14
#define BITS2	14
#define BITS3	4

#define MASK2	(NLEVEL2 - 1)
#define MASK3	(NLEVEL3 - 1)

/* Basic structure of a pte (page table entry). The user-defined pte
 * may contain any number of fields but the first field must be the
 * address.
 */
typedef struct pt_basic {
    unsigned int addr;
} pt_t, *pt_ptr;

#ifdef __cplusplus
extern "C" {
#endif

pt_t * pt_insert(pt_t ***level1, unsigned int addr, int size);
pt_t * pt_lookup(pt_t ***level1, unsigned int addr, int size);
void pt_apply(pt_t ***level1, PFI func, int size);
int pt_free(pt_t ***level1, int size);
int pt_space(pt_t ***level1, int size);

#ifdef  __cplusplus
}
#endif /* __cplusplus */

#endif /* !__pt_h */
