/*
 * Definition of the event structure and bit field mnemonics.
 *
 * Copyright (C) 1993 by Jack E. Veenstra (veenstra@cs.rochester.edu)
 * 
 * This file is part of MINT, a MIPS code interpreter and event generator
 * for parallel programs.
 * 
 * MINT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 * 
 * MINT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with MINT; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __task_h
#define __task_h

#include "event.h"
#include "export.h"

/* The order of the first few fields is important. The time wheel
 * defines a struct that contains only next, prev, time, and priority.
 */
typedef struct task {
    struct task *next;    /* forward link for task queue */
    struct task *prev;    /* backward link for task queue */
    mint_time_t time;	  /* absolute time at which to execute this task */
    int priority;	  /* larger numbers = higher priority */
    int pid;		  /* processor id */
    int (*ufunc)(struct task *);	  /* user-defined function to call */
    struct task *tstack;  /* a stack of tasks to execute next */
    struct event *pevent; /* pointer to an event structure */
    struct task *dbg_next; /* forward link for debug queue */
    struct task *dbg_prev; /* backward link for debug queue */
    int ival1;		  /* user-defined integer value */
    int ival2;		  /* user-defined integer value */
    int ival3;		  /* user-defined integer value */
    int ival4;		  /* user-defined integer value */
    int ival5;		  /* user-defined integer value */
    int ival6;		  /* user-defined integer value */
    int ival7;		  /* user-defined integer value */
    int ival8;		  /* user-defined integer value */
    int ival9;		  /* user-defined integer value */
    int ival10;		  /* user-defined integer value */
    int ival11;		  /* user-defined integer value */
    int ival12;		  /* user-defined integer value */
    double dval1;	  /* user-defined double value */
    double dval2;	  /* user-defined double value */
    mint_time_t tval1;	  /* user-defined time value */
    mint_time_t tval2;	  /* user-defined time value */
    mint_time_t tval3;	  /* user-defined time value */
    void *uptr;		  /* user-defined pointer */
    void *uptr2;	  /* user-defined pointer */
    void *uptr3;	  /* user-defined pointer */
    void *uptr4;	  /* user-defined pointer */
    int (*ufunc2)(struct task *);	  /* user-defined function pointer */
    struct task *unext;	  /* user-defined forward link */
    struct task *uprev;   /* user-defined backward link */
} task_t, *task_ptr;

typedef int (*PFTASK)(struct task *ptask);

/* the default priority */
#define DEF_PRIORITY 0

#define PRIORITY_HIGH (~(1 << 31))
#define PRIORITY_LOW (1 << 31)

#define T_ADVANCE	0
#define T_FREE		1
#define T_NO_ADVANCE	T_FREE	/* for backwards compatibility */
#define T_YIELD		2
#define T_CONTINUE	3
#define T_MAX		4	/* must be last */

#ifdef __cplusplus
extern "C" {
#endif

/* function prototypes */
int sim_init(int argc, char **argv);
void sim_done(double elapsed, double cpu_time);

int sim_read(task_ptr ptask);
int sim_write(task_ptr ptask);
int sim_user(task_ptr ptask);
int sim_terminate(task_ptr ptask);
int sim_exists(int pid);
int sim_prefork(task_ptr ptask);
void sim_fork(task_ptr ptask);
void sim_block(task_ptr ptask);
void sim_unblock(task_ptr ptask);
int sim_lock_attempt(task_ptr ptask);
int sim_lock_acquire(task_ptr ptask);
int sim_lock_release(task_ptr ptask);
int sim_barrier_attempt(task_ptr ptask);
int sim_barrier_acquire(task_ptr ptask);
int sim_psema_attempt(task_ptr ptask);
int sim_psema_acquire(task_ptr ptask);
int sim_vsema(task_ptr ptask);
int sim_alloc(task_ptr ptask);
int sim_shalloc(task_ptr ptask);
int sim_lockalloc(task_ptr ptask);
void sim_checkpoint(task_ptr ptask);

void task_destroy(int pid);
task_ptr sched_task(task_ptr ptask, PFTASK ufunc, mint_time_t tval);
task_ptr pri_sched_task(task_ptr ptask, PFTASK ufunc, mint_time_t tval,
                        int priority);
task_ptr sched_task_before(task_ptr ptask, task_ptr before, PFTASK ufunc);
task_ptr stack_task(task_ptr ptask, PFTASK ufunc);
task_ptr pri_stack_task(task_ptr ptask, PFTASK ufunc, int priority);
task_ptr schedule_task(task_ptr pnew);
task_ptr unblock_proc(int pid, mint_time_t tval);
task_ptr unblock_proc_now(int pid);
void cancel_task(task_ptr ptask);
task_ptr newtask();
void free_task(task_ptr ptask);

void debug_task(task_ptr ptask);

#ifdef  __cplusplus
}
#endif /* __cplusplus */

#endif /* !__task_h */
