/*
 * Definitions of all global variables that are used in more than one file,
 * not including globals that are exported to the simulator.
 *
 * Copyright (C) 1993 by Jack E. Veenstra (veenstra@cs.rochester.edu)
 * 
 * This file is part of MINT, a MIPS code interpreter and event generator
 * for parallel programs.
 * 
 * MINT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 * 
 * MINT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with MINT; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef EXTERN
#undef EXTERN
#endif

#ifdef MAIN
#define EXTERN
#else
#define EXTERN extern
#endif

/*
 * ROUNDUP() is a macro that rounds "val" up to a multiple of "align"
 * where "align" must be a power of two.
 */
#define ROUNDUP(val,align) (((val) + ((align) - 1)) & ~((align) - 1))

/*
 * Align the simulated address spaces on an 8K boundary.
 */
#define M_ALIGN (8 * 1024)

/*
 * Lockpage_size is the size of a page from which to allocate locks,
 * barriers, and semaphores. If this variable is zero, then locks and such
 * are allocated inline with other shared data. The default is 4096,
 * that is, locks are allocated in separate 4K pages from the rest of
 * shared data.
 */
EXTERN int Lockpage_size;
#define LOCKPAGE_SIZE 4096

/*
 * Lockblk_size is the size of a block (in bytes) in which to place each lock,
 * barrier, or semaphore. If non-zero, then each lock is allocated in
 * a separate cache block from other shared data and other locks.
 * If zero, then locks and such are allocated in the next free location,
 * and may share a cache block with other locks or data. The default is 0.
 */
EXTERN int Lockblk_size;
#define LOCKBLK_SIZE 0

EXTERN long Lkmem_start;
EXTERN long Lkmem_end;

/*
 * File descriptor for MINT output. Defaults to stderr. Output can be
 * turned off from the user program by setting this to NULL.
 */
EXTERN FILE *Mint_output;

/* File name for logging upshot-compatible events. */
EXTERN char *Upshot_log;

/* the number of existing processes */
EXTERN int Nprocs;

/* The file from which to generate fake events, if non-NULL */
EXTERN char *Fevents;

/* name of simulator executable */
EXTERN char *Simname;

/*
 * All the xxx_start variables are addresses in object's memory space,
 * except for Private_start.
 */
EXTERN long Private_start;	/* start of private memory */
EXTERN long Heap_start;		/* start of heap */
EXTERN long Shmem_start;	/* start of shared memory */
EXTERN long Shmem_end;		/* shared memory allocated so far */

EXTERN long Stack_size;		/* stack size in bytes */
EXTERN long Heap_size;		/* heap size in bytes */
EXTERN long Mem_size;		/* stack size in bytes */
EXTERN long Shmem_size;		/* shared memory size */

EXTERN int Mem_size_specified;	/* true if the "-m" option was used */

EXTERN event_t *pcur_event;	/* The current event */

/* allow changing these defaults on the "make" command line */
#ifndef SHMEM_SIZE
#ifdef sgi
#define SHMEM_SIZE (16 * 1024 * 1024)	/* shared memory size */
#else
/* use a smaller default for machines with less memory */
#define SHMEM_SIZE (0x00ffffff)	/* shared memory size */
#endif
#endif
#ifndef PRIVATE_START
#define PRIVATE_START 0x20000000
#endif
#ifndef HEAP_START
#define HEAP_START    0x30000000
#endif
#ifndef SHARED_START
#define SHARED_START  0x40000000
#endif

/* default stack size */
#ifndef STACK_SIZE
#define STACK_SIZE (16 * 1024)
#endif

/* default heap size */
#ifndef HEAP_SIZE
#define HEAP_SIZE (256 * 1024)
#endif

/* use this to enforce a minimum size for the stack */
#ifndef STACK_SIZE_MIN
#define STACK_SIZE_MIN (8 * 1024)
#endif

/* use this to enforce a minimum size for the heap */
#ifndef HEAP_SIZE_MIN
#define HEAP_SIZE_MIN (8 * 1024)
#endif

#ifndef DEF_MAX_NPROCS

#ifdef sgi
#define DEF_MAX_NPROCS 256
#else
/* use a smaller default for machines with less memory */
#define DEF_MAX_NPROCS 32
#endif

#endif

/*
 * Ckpoint_freq controls the frequency of calls to sim_checkpoint().
 * This can be set on the command line as well as changed by the user while
 * the program is running. If set to zero, no further calls to sim_checkpoint()
 * will occur.
 */
#define DEFAULT_CKPOINT_FREQ 10000000
EXTERN unsigned int Ckpoint_freq;

/*
 * All queues are circular doubly linked lists with a head node.
 * When a thread exits, it goes on the Done_q until the parent waits on
 * it. If a parent exits before its children, it goes to sleep. When the
 * child exits, it wakes up its parent. The parent moves finished children
 * from the Done_q to the Free_q.
 *
 * Run_q is the head node for the run queue.
 * Free_q is the head node for the free queue.
 *  ... and so on.
 */
EXTERN thread_t Run_q;
EXTERN thread_t Free_q;
EXTERN thread_t Done_q;
EXTERN thread_t Sleep_q;

/* the array of threads, indexed by pid */
EXTERN thread_ptr Threads;

/* Free list pointer for events. One event is allocated for each thread. */
EXTERN event_ptr Event_free;

/* the maximum process id ever used (process ids start at zero) */
EXTERN int Maxpid;

/*
 * The maximum number of processes that can be simulated. This is used
 * to allocate space. This value can be changed with the "-p" command
 * line option.
 */
EXTERN int Max_nprocs;

/* pointer to array of free semaphores */
EXTERN sema_ptr Sema_free;

/* head node for storing list of allocated semaphores */
EXTERN sema_t Sema_list;

/* pointer to array of free locks */
EXTERN uslock_ptr Lock_free;

/* head node for storing list of allocated locks */
EXTERN uslock_t Lock_list;

/* globals set by command line options */
EXTERN int Mint_debug;		/* also requires compiling with -DTRACE */
EXTERN int Verify_protocol;
EXTERN int Locks_block;		/* spinlocks block instead of spinning */
EXTERN int Recycle_threads;	/* reuse pid of a dead child on next fork */
EXTERN int Quiet;

/* The time of the first fork, or zero if no forks yet */
EXTERN mint_time_t First_fork_elapsed;

/* The accumulated cpu time of processor 0 at the time of the first fork */
EXTERN mint_time_t First_fork_cpu;

/* Pointer to the simulator's context */
EXTERN thread_ptr psim_thread;

/* Pointer to the thread structure of the current application thread */
EXTERN thread_ptr pcur_thread;

/* function prototypes */
void block_thr (qnode_ptr q, thread_ptr pthread);
thread_ptr unblock_thr (thread_ptr q);
void init_child_addr_space (thread_ptr parent, thread_ptr child);
int wait_for_children (thread_ptr pthread);
int wait_for_child (thread_ptr pthread);
int task_yield (task_ptr ptask);
int task_exit (task_ptr ptask);
int task_done (task_ptr ptask);
int task_fork (task_ptr ptask);
int task_block (task_ptr ptask);
int task_unblock (task_ptr ptask);
int task_unblock_list (task_ptr ptask);
int task_terminate (task_ptr ptask);
int task_synch_time (task_ptr ptask);

/* main() of the application */
int appl_main (int argc, char **argv);

/* argc and argv for appl_main */
EXTERN int appl_argc;
EXTERN char **appl_argv;

/* either next_event or next_fake_event */
EXTERN void (*ctxsw_fn) (thread_ptr ptr1, thread_ptr ptr2, thread_ptr * ptr3);
#ifdef INTEL
void ctxsw_intel (thread_ptr ptr1, thread_ptr ptr2, thread_ptr * ptr3);
#else /* INTEL */
void ctxsw (thread_ptr ptr1, thread_ptr ptr2, thread_ptr * ptr3);
#endif /* INTEL */
void fake_ctxsw (thread_ptr dummy1, thread_ptr pthread, thread_ptr * dummy2);

typedef struct fake_event_t {
    int pid;
    mint_time_t time;
    int addr;
    int type;
    int utype;
    int next;
} fake_event_t, *fake_event_ptr;

/* Pointer to the fake event structure */
EXTERN fake_event_ptr fevent;

/* global system time */
EXTERN mint_time_t cur_time;

/*
 * Since fiaddl works only with a memory location and not an immediate
 * constant, we needed this variable - arsharma */
EXTERN int static_execution_time;
