/*
 * Routines to implement a "page table" data structure. Given a 32-bit
 * address, it returns the corresponding user-defined structure.
 *
 * Copyright (C) 1993 by Jack E. Veenstra (veenstra@cs.rochester.edu)
 * 
 * This file is part of MINT, a MIPS code interpreter and event generator
 * for parallel programs.
 * 
 * MINT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 * 
 * MINT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with MINT; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Change log:
 *
 * Sep  7, 1992         Jack Veenstra           veenstra@cs.rochester.edu
 *      Created.
 *
 * Mar 16, 1993         Jack Veenstra           veenstra@cs.rochester.edu
 *      Fixed a bug that allocated too much space for the second level
 * pointers.
 *
 */

#include <stdio.h>
#include "pt.h"

/*
 * These routines implement a "page table" data structure. Each routine uses
 * "level1" as the top-level pointer to a tree of page table entries. The
 * routines pt_lookup() and pt_insert search for the address "addr" by
 * using the bits in addr to index the page table. If the
 * corresponding page table entry is found, these routines return a
 * pointer to the user-defined structure "struct pte". If not found,
 * pt_lookup() returns NULL; pt_insert() allocates a new entry and
 * returns a pointer to it. 
 *
 * The program should declare and define an array of pointers to pointers
 * to "pt_t" to pass as the first argument to these routines.
 *     Example:
 *
 *     pt_t **cache[NLEVEL1];
 *
 * or just declare the pointer and call "pt_init()" which will allocate
 * the space:
 *
 *     pt_t ***cache;
 *     cache = pt_init();
 *
 * The program using the page table routines should define a structure
 * for containing page table information. The struct must include as its
 * first field an unsigned int field to store the address. Other fields
 * are optional. The size of the structure must be passed as an argument
 * so a macro definition for the size is useful.
 *
 * Example:
 *
 * struct pte {
 *     unsigned int addr;
 *     unsigned int accessed;
 *     unsigned int written;
 * };
 *
 * #define PTE_SIZE (sizeof(struct pte))
 *
 * Then the call to translate an address (and insert it if not found)
 * would look like:
 *
 *     ptep = pt_insert(cache, addr, PTE_SIZE);
 *
 * To lookup an address without inserting it if it's not found:
 *
 *     ptep = pt_lookup(cache, addr, PTE_SIZE);
 */

pt_t ***
pt_init ()
{
    pt_t ***level1;

    level1 = (pt_t ***) calloc (NLEVEL1, sizeof (pt_t *));
    if (level1 == NULL)
	fatal ("pt_init: out of memory, allocating %u pointers\n", NLEVEL1);
    return level1;
}

pt_t *
pt_insert (level1, addr, size)
     pt_t **level1[];
     unsigned int addr;
     int size;
{
    int index1, index2, index3;
    pt_t **level2;
    pt_t *ptep;
    char *level3;

    index1 = addr >> (32 - BITS1);
    if (level1[index1] == NULL) {
	level1[index1] = (pt_t **) calloc (NLEVEL2, sizeof (pt_t *));
	if (level1[index1] == NULL) {
	    fprintf (stderr,
                     "pt_insert: out of memory (level1) addr=%x\n",
                     addr);
	    exit (1);
	}
    }

    level2 = level1[index1];
    index2 = (addr >> (32 - BITS1 - BITS2)) & MASK2;
    if (level2[index2] == NULL) {
	level2[index2] = (pt_t *) calloc (NLEVEL3, size);
	if (level2[index2] == NULL) {
	    fprintf (stderr,
                     "pt_insert: out of memory (level2) addr=%x\n",
                     addr);
	    exit (1);
	}
    }

    level3 = (char *) level2[index2];
    index3 = (addr >> (32 - BITS1 - BITS2 - BITS3)) & MASK3;
    ptep = (pt_t *) & level3[index3 * size];
    ptep->addr = addr;
    return ptep;
}

pt_t *
pt_lookup (level1, addr, size)
     pt_t **level1[];
     unsigned int addr;
     int size;
{
    int index1, index2, index3;
    pt_t **level2;
    pt_t *ptep;
    char *level3;

    index1 = addr >> (32 - BITS1);
    if (level1[index1] == NULL)
	return NULL;

    level2 = level1[index1];
    index2 = (addr >> (32 - BITS1 - BITS2)) & MASK2;
    if (level2[index2] == NULL)
	return NULL;

    level3 = (char *) level2[index2];
    index3 = (addr >> (32 - BITS1 - BITS2 - BITS3)) & MASK3;
    ptep = (pt_t *) & level3[index3 * size];
    if (ptep->addr == (unsigned int) NULL)
	return NULL;
    if (ptep->addr != addr) {
	fprintf (stderr, "pt_lookup: ERROR: ptep->addr (0x%x) "
                 "does not match addr (0x%x)\n",
		 ptep->addr, addr);
	exit (1);
    }
    return ptep;
}

/*
 * pt_apply() calls func(), a user-supplied function, for every entry
 * in the page table. A pointer to a page table entry is passed to the
 * function. The entries are passed to the user function in numerically
 * increasing order. The user function should return TRUE (1) in order
 * to continue processing entries. To "break out of the loop" and
 * discontinue processing entries, the user function should return
 * FALSE (0).
 */

void
pt_apply (level1, func, size)
     pt_t **level1[];
     PFI func;
     int size;
{
    int index1, index2, index3;
    pt_t **level2;
    pt_t *ptep;
    char *level3;
    int last3;

    last3 = NLEVEL3 * size;
    for (index1 = 0; index1 < NLEVEL1; index1++) {
	if (level1[index1] == NULL)
	    continue;
	level2 = level1[index1];
	for (index2 = 0; index2 < NLEVEL2; index2++) {
	    if (level2[index2] == NULL)
		continue;
	    level3 = (char *) level2[index2];
	    for (index3 = 0; index3 < last3; index3 += size) {
		ptep = (pt_t *) & level3[index3];
		if (ptep->addr != (unsigned int) NULL)
		    if ((*func) (ptep) == 0)
			return;
	    }
	}
    }
}

/*
 * pt_free() frees the space used by the page table entries and returns
 * the amount of space used.
 */

int
pt_free (level1, size)
     pt_t **level1[];
     int size;
{
    int index1, index2;
    pt_t **level2;
    int space;

    space = 0;
    for (index1 = 0; index1 < NLEVEL1; index1++) {
	if (level1[index1] == NULL)
	    continue;
	level2 = level1[index1];
	for (index2 = 0; index2 < NLEVEL2; index2++) {
	    if (level2[index2]) {
		free (level2[index2]);
		space += NLEVEL3 * size;
	    }
	}
	free (level2);
	space += NLEVEL2 * sizeof (pt_t *);
    }
    space += NLEVEL1 * sizeof (pt_t **);
    return space;
}

/*
 * pt_space() returns the amount of space used in the page table.
 */

int
pt_space (level1, size)
     pt_t **level1[];
     int size;
{
    int index1, index2;
    pt_t **level2;
    int space, space3;

    space = 0;
    space3 = NLEVEL3 * size;
    for (index1 = 0; index1 < NLEVEL1; index1++) {
	if (level1[index1] == NULL)
	    continue;
	level2 = level1[index1];
	for (index2 = 0; index2 < NLEVEL2; index2++) {
	    if (level2[index2]) {
		space += space3;
	    }
	}
	space += NLEVEL2 * sizeof (pt_t *);
    }
    space += NLEVEL1 * sizeof (pt_t **);
    return space;
}
