
#include <stdio.h>
#include "icode.h"
#include "globals.h"
#include "task.h"

#define BUFSIZE 120
#define MAX_FAKE_EVENTS  0x100000	/* 1M events */

typedef struct tname_t {
    char *name;
    int len;
    int type;
    PFTASK func;
} tname_t, *tname_ptr;

static tname_t Typenames[] =
{
    "read", 1, E_READ | E_SHARED | E_READ_OFFSET, sim_read,
    "write", 1, E_WRITE | E_SHARED | E_WRITE_OFFSET, sim_write,
    "user", 1, E_USER_EVENT, sim_user,
    "lock_attempt", 7, E_LOCK_ATTEMPT, sim_lock_attempt,
    "lock_acquire", 7, E_LOCK_ACQUIRE, sim_lock_acquire,
    "lock_release", 6, E_LOCK_RELEASE, sim_lock_release,
    "barrier_attempt", 10, E_BARRIER_ATTEMPT, sim_barrier_attempt,
    "barrier_acquire", 10, E_BARRIER_ACQUIRE, sim_barrier_acquire,
    "psema_attempt", 8, E_PSEMA_ATTEMPT, sim_psema_attempt,
    "psema_acquire", 8, E_PSEMA_ACQUIRE, sim_psema_acquire,
    "vsema", 1, E_VSEMA, sim_vsema,
    NULL, 0, 0, NULL
};

void error_at (char *fname, int lineno, char *line, char *ptr, char *errmsg);
void print_pointer (char *ptr, char *line);

void
init_fake_threads ()
{
    int i;
    thread_ptr pthread;
    event_ptr pevent;

    Maxpid = Max_nprocs - 1;
    printf ("Max_nprocs = %d\n", Max_nprocs);
    Nprocs = Max_nprocs;
    for (i = 0; i < Max_nprocs; i++) {
	pthread = &Threads[i];
	INLINE_REMOVE (pthread);	/* remove thread from Free_q */
	NEW_ITEM (Event_free, sizeof (event_t), pevent, "init_fake_thread");
	init_thread (pthread, pevent);
	pthread->time = 0;
	pthread->parent = NULL;
	pthread->sibling = NULL;
	pthread->fcr31 = 0;
	/*
	 *
	 * The index to the first event will be set in read_events
	 * if the index is -1 the thread terminates
	 */
	pthread->index = -1;
    }
}

read_events (char *fname)
{
    int i, pid, simtime, data_addr, type, utype, next, len;
    FILE *fd;
    char *cp, *endptr, *dest;
    char buf[BUFSIZE], typename[BUFSIZE];
    int lineno;
    tname_ptr ptype;
    int prev[MAXPROC];
    thread_ptr pthread;
    int index;

    /* open the file */
    fd = fopen (fname, "r");
    if (fd == NULL) {
	perror (fname);
	exit (1);
    }

    /* initialize the fake event structure */
    fevent = (fake_event_ptr) malloc (MAX_FAKE_EVENTS * sizeof (fake_event_t));
    if (fevent == NULL)
	fatal ("read_events: cannot allocate memory for fevent\n");
    index = 0;

    /*
     * Read the file. Each line contains the following fields:
     * pid time data_addr type utype
     */
    for (lineno = 1; cp = fgets (buf, BUFSIZE, fd); lineno++) {
	if (*cp == '#' || *cp == '\n')
	    continue;

	pid = strtol (cp, &endptr, 0);
	if (pid >= Max_nprocs) {
	    fprintf (stderr, "read_events: pid (%d) greater than max (%d).\n",
		     pid, Max_nprocs - 1);
	    error_at (fname, lineno, buf, cp, "bad pid value");
	}
	if (cp == endptr)
	    error_at (fname, lineno, buf, cp, "bad pid value");
	cp = endptr;

	simtime = strtol (cp, &endptr, 0);
	if (cp == endptr)
	    error_at (fname, lineno, buf, cp, "bad time value");
	cp = endptr;

	data_addr = strtol (cp, &endptr, 0);
	if (cp == endptr)
	    error_at (fname, lineno, buf, cp, "bad data address");
	cp = endptr;

	/* read the type name */
	while (isspace (*cp))
	    cp++;
	dest = typename;
	while (!isspace (*cp))
	    *dest++ = *cp++;
	*dest = 0;
	/* convert to lower case */
	for (dest = typename; *dest; dest++)
	    if (*dest >= 'A' && *dest <= 'Z')
		*dest += 'a' - 'A';
	for (ptype = Typenames; ptype->name; ptype++) {
	    len = strlen (typename);
	    if (ptype->len > len)
		len = ptype->len;
	    if (strncmp (ptype->name, typename, len) == 0) {
		type = ptype->type;
		break;
	    }
	}
	if (ptype->name == NULL)
	    error_at (fname, lineno, buf, cp, "unknown type");
	else if (type == E_USER_EVENT) {
	    utype = strtol (cp, &endptr, 0);
	    if (cp == endptr)
		error_at (fname, lineno, buf, cp, "bad user type");
	    cp = endptr;
	}

	if (index >= MAX_FAKE_EVENTS)
	    fatal ("read_events: max number of fake events "
                   "exceeded %d\n", index);
	printf ("pid = %d pth->ind %d prev %d\n",
		pid, Threads[pid].index, prev[pid]);
	if (Threads[pid].index == -1) {
	    Threads[pid].index = index;
	    prev[pid] = index;
	}
	else {
	    fevent[prev[pid]].next = index;
	}
	fevent[index].pid = pid;
	fevent[index].time = simtime;
	fevent[index].addr = data_addr;
	fevent[index].type = type;
	fevent[index].utype = utype;
	fevent[index].next = -1;
	printf ("read_events: ind %d pid %d time %d tp 0x%x %d\n",
		index, pid, simtime, type, utype);
	prev[pid] = index;
	index++;
    }
    fclose (fd);
    for (pid = 0; pid < Nprocs; pid++) {
	if (Threads[pid].index == -1) {
	    Threads[pid].index = index;
	    prev[pid] = index;
	}
	else {
	    fevent[prev[pid]].next = index;
	}
	fevent[index].pid = pid;
	fevent[index].time = 0;
	fevent[index].type = E_CHECK_TASKS;
	fevent[index].next = index + 1;
	prev[pid] = index;
	index++;
	fevent[index].pid = pid;
	fevent[index].time = 0;
	fevent[index].type = E_DONE_THR;
	fevent[index].next = -1;
	prev[pid] = index;
	index++;
    }
}

/* indent the display of the line that caused the error */
#define ERROR_INDENT "    "

/*
 * Prints filename, line number, and error message, with a pointer to
 * the place in the line where the error was detected.
 */
void
error_at (char *fname, int lineno, char *line, char *ptr, char *errmsg)
{
    printf ("Error in file `%s' at line %d: %s\n%s%s",
	    fname, lineno, errmsg, ERROR_INDENT, line);
    if (line[strlen (line) - 1] != '\n')
	printf ("\n");
    print_pointer (ptr, line);
    exit (1);
}

void
print_pointer (char *ptr, char *line)
{
    int i, nchars, nspaces, space_left;
    static char Hyphens[81] = "---------------------"
        "-----------------------------------------------------------";

    nspaces = 0;
    nchars = ptr - line;
    for (i = 0; i < nchars; i++) {
	nspaces++;
	if (line[i] == '\t') {
	    space_left = nspaces % 8;
	    nspaces += space_left ? 8 - space_left : 0;
	}
    }
    printf ("%s%s^\n", ERROR_INDENT, &Hyphens[80 - nspaces]);
}
