#include <stdio.h>
#include "icode.h"
#include "globals.h"
#include "task.h"
#include "wheel.h"

/*
 * The code that uses "unsigned long" for storing time values is inside
 * ifdef ULONG_TIME. This code will probably be unsupported in the future.
 */

/*
 * This points to the current position within the time wheel. We used to
 * be able to use inline macros and make this a local variable but we
 * can't anymore because task_insert_current() needs to access this
 * pointer at arbitrary times.
 */
wheel_task_ptr Curptr;

/*
 * The time wheel is an array of head nodes to doubly-linked lists.
 * The array is indexed by the time value of a task. If the time
 * value is too large, then a task is added to a "Future queue"
 * and processed when the current index into the time wheel wraps around.
 * Curptr is a pointer to the head node for the current time wheel entry.
 * When all tasks for the current time have been executed, Curptr is
 * incremented to the next time wheel entry.
 */
wheel_task_ptr Wheel;

/*
 * Endptr points past the end of the time wheel. Once initialized, this
 * variable never changes.
 */
wheel_task_ptr Endptr;

/* Start_time is only used when the type of mint_time_t is a "double". */
mint_time_t Start_time;

/*
 * End_time is the largest time value that can be used to index the
 * time wheel. End_time changes when the time wheel wraps around.
 */
mint_time_t End_time;

/* Future_q stores tasks that are scheduled for too far in the future. */
task_ptr Future_q;
static task_t Future_q_head;
static task_t Dummy_task;

/* used in wheel_print() to lookup function names */
/* extern hash_tab_ptr Func_hash_tab; */
extern int Num_tasks;

/* if more than this number of tasks are allocated, it's probably a mistake */
#define TOO_MANY_TASKS 10000

/* use a global variable so that it can be changed by the back-end */
int Too_many_tasks = TOO_MANY_TASKS;

int ckpoint_event (task_ptr ptask);

void
wheel_init ()
{
    int i;

    Start_time = 0;
    End_time = WHEEL_SIZE - 1;
    Wheel = (wheel_task_ptr) calloc (sizeof (wheel_task_t), WHEEL_SIZE + 1);
    if (Wheel == NULL)
	fatal ("wheel_init: cannot allocate %d bytes for the time wheel.\n",
	       sizeof (wheel_task_t) * (WHEEL_SIZE + 1));
    for (i = 0; i < WHEEL_SIZE + 1; i++) {
	INLINE_INIT_Q ((task_ptr) & Wheel[i]);
	Wheel[i].priority = PRIORITY_HIGH;
	Wheel[i].time = MAXTIME;
    }
    Curptr = &Wheel[0];
    Endptr = &Wheel[WHEEL_SIZE];
    /* add a dummy item to the last (unused) queue in the time wheel */
    INLINE_ENQUEUE ((task_ptr) & Wheel[WHEEL_SIZE], &Dummy_task);
    Future_q = &Future_q_head;
    INLINE_INIT_Q (Future_q);
    Future_q->time = MAXTIME;
    Dummy_task.pid = -2;	/* an unused pid */
    Future_q_head.pid = -2;	/* an unused pid */
    if (Ckpoint_freq > 0)
	pri_sched_task (&Dummy_task, ckpoint_event,
                        Ckpoint_freq, PRIORITY_LOW);
}

ckpoint_event (task_ptr ptask)
{
#ifdef DEBUG
    static already_printed = 0;
#endif

#ifndef CALL_RUNNET
    int i;

    /* if there are no scheduled tasks then all are blocked */
    for (i = 0; i < WHEEL_SIZE; i++)
	if (Wheel[i].next != (task_ptr) & Wheel[i])
	    break;
    if (i == WHEEL_SIZE) {
	/* check the future queue */
	if (Future_q->next == Future_q)
	    fatal ("no more scheduled tasks to run;"
                   " processes are deadlocked.\n");
    }
#endif

#ifdef DEBUG
    if (Num_tasks > Too_many_tasks && !already_printed) {
	/* print this warning only once to avoid infinite IO */
	already_printed = 1;
	fprintf (stderr, "\nThere are %d tasks in use "
                 "-- could be a memory leak.\n",
		 Num_tasks);
	fprintf (stderr, "The first 20 tasks are:\n");
	print_active_tasks (20);
	fprintf (stderr, "\nTo print the whole list, call "
                 "\"print_active_tasks(0)\"\n\n");
    }
#endif
    sim_checkpoint (ptask);
    if (Ckpoint_freq > 0)
	pri_sched_task (&Dummy_task, ckpoint_event,
			ptask->time + Ckpoint_freq, PRIORITY_LOW);
    return T_NO_ADVANCE;
}

/*
 * Returns true if time wheel is empty (not counting the checkpoint task)
 */
wheel_empty ()
{
    int i;
    task_ptr pnew;

    for (i = 0; i < WHEEL_SIZE; i++) {
	pnew = Wheel[i].next;
	if (pnew->ufunc == ckpoint_event)
	    pnew = pnew->next;
	if (pnew != (task_ptr) & Wheel[i])
	    return 0;
    }
    pnew = Future_q->next;
    if (pnew->ufunc == ckpoint_event)
	pnew = pnew->next;
    if (pnew != Future_q)
	return 0;
    return 1;
}

/*
 * Returns true if the time wheel contains no tasks for thread pid.
 */
wheel_empty_pid (int pid)
{
    int i;
    task_ptr pnew;

    for (i = 0; i < WHEEL_SIZE; i++) {
	pnew = Wheel[i].next;
	while (pnew != (task_ptr) & Wheel[i]) {
	    if (pid == pnew->pid)
		return 0;
	    pnew = pnew->next;
	}
    }
    pnew = Future_q->next;
    while (pnew != Future_q) {
	if (pid == pnew->pid)
	    return 0;
	pnew = pnew->next;
    }
    return 1;
}

/*
 * task_cleanup() ensures that all the scheduled tasks for a process
 * have executed before sim_terminate() is called.
 */
task_cleanup (task_ptr ptask)
{
    int i;
    task_ptr pnew;

    /* search the Future queue first */
    for (pnew = Future_q->next; pnew != Future_q; pnew = pnew->next)
	if (pnew->pid == ptask->pid)
	    break;

    /*
     * If we found a task for this process, wait for it to execute, then
     * check the wheel again at that time.
     */
    if (pnew != Future_q) {
	pri_sched_task (ptask, task_cleanup, pnew->time, PRIORITY_LOW);
#ifdef DEBUG
	warning ("process %d exited before task at time %.0f executed.\n",
		 ptask->pid, (double) pnew->time);
#endif
	return T_NO_ADVANCE;
    }

    /* search backwards, starting at the end of the wheel */
    for (i = WHEEL_SIZE - 1; i >= 0; i--) {
	pnew = Wheel[i].prev;
	/* if the queue is empty, continue with the next queue */
	if (pnew == (task_ptr) & Wheel[i])
	    continue;
	while (pnew != (task_ptr) & Wheel[i]) {
	    if (pnew->pid == ptask->pid)
		break;
	    pnew = pnew->prev;
	}

	/*
	 * If we found a task for this process, wait for it to execute, then
	 * check the wheel again at that time.
	 */
	if (pnew != (task_ptr) & Wheel[i]) {
	    pri_sched_task (ptask, task_cleanup, pnew->time, PRIORITY_LOW);
#ifdef DEBUG
	    warning ("process %d exited before task at time %.0f executed.\n",
		     ptask->pid, (double) pnew->time);
#endif
	    return T_NO_ADVANCE;
	}
    }
    return T_CONTINUE;
}

/*
 * This function is only called when DEBUG is used. Otherwise an inline
 * macro call in the execution loop is used.
 */
#if defined(DEBUG) || defined(NO_INLINING)
#ifdef ULONG_TIME
task_ptr
task_extract ()
{
    task_ptr pnew, tptr, next;
    unsigned long index, itime;

    /* search the time wheel first */
    pnew = Curptr->next;
    if (pnew != (task_ptr) Curptr) {
	INLINE_REMOVE (pnew);
	return pnew;
    }

    while (1) {

	/*
	 * Search the queues until we get to the end. Since we put
	 * a dummy item in the last (unused) queue, we don't need
	 * to check for overflow of the Curptr in each iteration.
	 */
	do {
	    Curptr++;
	    pnew = Curptr->next;
	} while (pnew == (task_ptr) Curptr);

	/* if we found a task, remove it and return */
	if (Curptr != Endptr) {
	    INLINE_REMOVE (pnew);
	    return pnew;
	}

	/* wrap around processing */
	Curptr = &Wheel[0];
	End_time += WHEEL_SIZE;
	pnew = Future_q->next;
	if (pnew == Future_q)
	    return NULL;

	for (; pnew != Future_q; pnew = next) {
	    next = pnew->next;
	    itime = pnew->time;
	    index = itime & WHEEL_MASK;

	    if (itime <= End_time) {
		/* insert in priority order */
		for (tptr = Wheel[index].prev; tptr->priority < pnew->priority;
		     tptr = tptr->prev);

		INLINE_REMOVE (pnew);
		INLINE_INSERT_AFTER (tptr, pnew);
	    }
	}

	pnew = Curptr->next;
	if (pnew != (task_ptr) Curptr) {
	    INLINE_REMOVE (pnew);
	    return pnew;
	}
    }
}

#else

task_ptr
task_extract ()
{
    task_ptr pnew, tptr, next;
    int diff;

    /* search the time wheel first */
    pnew = Curptr->next;
    if (pnew != (task_ptr) Curptr) {
	INLINE_REMOVE (pnew);
	return pnew;
    }

    while (1) {

	/*
	 * Search the queues until we get to the end. Since we put
	 * a dummy item in the last (unused) queue, we don't need
	 * to check for overflow of the Curptr in each iteration.
	 */
	do {
	    Curptr++;
	    pnew = Curptr->next;
	} while (pnew == (task_ptr) Curptr);

	/* if we found a task, remove it and return */
	if (Curptr != Endptr) {
	    INLINE_REMOVE (pnew);
	    return pnew;
	}

	/* wrap around processing */
	Curptr = &Wheel[0];
	End_time += WHEEL_SIZE;
	Start_time += WHEEL_SIZE;
	pnew = Future_q->next;
	if (pnew == Future_q)
	    return NULL;

	for (; pnew != Future_q; pnew = next) {
	    next = pnew->next;
	    diff = pnew->time - Start_time;
#ifdef DEBUG
	    if (diff < 0)
		fatal ("task_extract: what? diff negative? %lf %lf\n",
		       pnew->time, Start_time);
#endif

	    if (diff < WHEEL_SIZE) {
		/* insert in priority order */
		for (tptr = Wheel[diff].prev; tptr->priority < pnew->priority;
		     tptr = tptr->prev);

		INLINE_REMOVE (pnew);
		INLINE_INSERT_AFTER (tptr, pnew);
	    }
	}

	pnew = Curptr->next;
	if (pnew != (task_ptr) Curptr) {
	    INLINE_REMOVE (pnew);
	    return pnew;
	}
    }
}
#endif
#endif /* DEBUG */

/*
 * task_min() returns the task with the minimum time without removing
 * the task from the time wheel. If there are no tasks, it returns NULL.
 *   Apr 1993   Leonidas Kontothanassis (kthanasi@cs.rochester.edu)
 *      Created.
 *   May 1993   Jack Veenstra   (veenstra@cs.rochester.edu)
 *      Added the "if" test at the beginning to speed up the common case,
 *   and changed the array index lookup to use pointers.
 */
task_ptr
task_min ()
{
    task_ptr pnew, minp;
    wheel_task_ptr pwheel;
    mint_time_t min_time;

    /* search the time wheel first */
    pwheel = Curptr;
    pnew = pwheel->next;
    if (pnew != (task_ptr) pwheel)
	return pnew;

    /*
     * Search the queues until we get to the end. Since we put
     * a dummy item in the last (unused) queue, we don't need
     * to check for overflow in each iteration.
     */
    do {
	pwheel++;
	pnew = pwheel->next;
    } while (pnew == (task_ptr) pwheel);

    /* if we found a task, return it */
    if (pwheel != Endptr) {
	return pnew;
    }

    /* the time wheel is empty, so search the future queue */
    pnew = Future_q->next;
    if (pnew == Future_q)
	return NULL;

    /* wrap around processing */
    min_time = MAXTIME;
    minp = NULL;
    for (; pnew != Future_q; pnew = pnew->next) {
	if (pnew->time < min_time) {
	    min_time = pnew->time;
	    minp = pnew;
	}
    }
    return minp;
}

#ifdef ULONG_TIME
void
task_insert (task_ptr ptask)
{
    unsigned long index, itime;
    task_ptr tptr;

    itime = ptask->time;

    if (itime <= End_time) {
	index = itime & WHEEL_MASK;
	/* insert in priority order */
	for (tptr = Wheel[index].prev; tptr->priority < ptask->priority;
	     tptr = tptr->prev);

	INLINE_INSERT_AFTER (tptr, ptask);
    }
    else
	INLINE_ENQUEUE (Future_q, ptask);
}

#else

void
task_insert (task_ptr ptask)
{
    int index;
    task_ptr tptr;

    index = ptask->time - Start_time;
#ifdef DEBUG
    if (index < 0)
	fatal ("task_insert: what? diff negative? %lf %lf\n",
	       ptask->time, Start_time);
    if (index < Curptr - &Wheel[0])
	fatal ("task_insert: time of task, %f, precedes current time %f\n",
	       ptask->time, Start_time + (long) (Curptr - &Wheel[0]));
		   
#endif

    if (index < WHEEL_SIZE) {
	/* insert in priority order */
	for (tptr = Wheel[index].prev; tptr->priority < ptask->priority;
	     tptr = tptr->prev);

	INLINE_INSERT_AFTER (tptr, ptask);
    }
    else
	INLINE_ENQUEUE (Future_q, ptask);
}
#endif

void
task_insert_before (task_ptr ptask, task_ptr before)
{
#ifdef DEBUG
    int index;
#endif

#ifdef DEBUG
    index = ptask->time - Start_time;
    if (index < 0)
	fatal ("task_insert: what? diff negative? %lf %lf\n",
	       ptask->time, Start_time);
    if (index < Curptr - &Wheel[0])
	fatal ("task_insert: time of task, %f, precedes current time %f\n",
	       ptask->time, Start_time + (long) (Curptr - &Wheel[0]));
    if (before->time != ptask->time)
	fatal ("task_insert_before: task time (%.0f) != before time (%.0f)\n",
	       ptask->time, before->time);
#endif
    INLINE_INSERT_BEFORE (before, ptask);
}

/*
 * This routine inserts a task at the current time.
 */
void
task_insert_current (task_ptr ptask)
{
    task_ptr tptr;

    ptask->time = Start_time + (long) (Curptr - &Wheel[0]);

    /* insert in priority order */
    for (tptr = Curptr->prev; tptr->priority < ptask->priority;
	 tptr = tptr->prev);

    INLINE_INSERT_AFTER (tptr, ptask);
}

/*
 * This routine searches the time wheel slot for a task with a matching
 * function pointer "ptask->ufunc". If such a task is found, a pointer
 * to it is returned. Otherwise a NULL pointer is returned.
 */
task_ptr
already_scheduled (task_ptr ptask, mint_time_t simtime)
{
    int index;
    task_ptr tptr;
    int (*ufunc) ();

    index = simtime - Start_time;
#ifdef DEBUG
    if (index < 0)
	fatal ("already_scheduled: what? diff negative? %lf %lf\n",
	       ptask->time, Start_time);
    if (index < Curptr - &Wheel[0])
	fatal ("already_scheduled: time of task, %.1f, "
               "precedes current time %.1f\n",
	       ptask->time, Start_time + (long) (Curptr - &Wheel[0]));
#endif

    ufunc = ptask->ufunc;
    if (index < WHEEL_SIZE) {
	/* search the wheel entry queue for a matching function ponter */
	for (tptr = Wheel[index].next;
             tptr != (task_ptr) & Wheel[index];
             tptr = tptr->next)
	    if (tptr->ufunc == ufunc)
		return tptr;
    }
    else {
	/* search the future queue for a matching function ponter and time */
	for (tptr = Future_q->next; tptr != Future_q; tptr = tptr->next)
	    if (tptr->ufunc == ufunc && tptr->time == simtime)
		return tptr;
    }

    /* if we get here, then we did not find a matching function pointer */
    return NULL;
}

/*
 * This routine searches the time wheel slot for a task with a matching
 * function pointer "ptask->ufunc" and pid. If such a task is found, a pointer
 * to it is returned. Otherwise a NULL pointer is returned.
 */
task_ptr
already_scheduled_pid (task_ptr ptask, int pid, mint_time_t simtime)
{
    int index;
    task_ptr tptr;
    int (*ufunc) ();

    index = simtime - Start_time;
#ifdef DEBUG
    if (index < 0)
	fatal ("already_scheduled: what? diff negative? %lf %lf\n",
	       ptask->time, Start_time);
    if (index < Curptr - &Wheel[0])
	fatal ("already_scheduled: time of task, %.1f, "
               "precedes current time %.1f\n",
	       ptask->time, Start_time + (long) (Curptr - &Wheel[0]));
#endif

    ufunc = ptask->ufunc;
    if (index < WHEEL_SIZE) {
	/*
	 * Search the wheel entry queue for a matching function ponter
	 * and pid.
	 */
	for (tptr = Wheel[index].next;
             tptr != (task_ptr) & Wheel[index];
             tptr = tptr->next)
	    if (tptr->ufunc == ufunc && tptr->pid == pid)
		return tptr;
    }
    else {
	/*
	 * Search the future queue for a matching function ponter,
	 * pid and time.
	 */
	for (tptr = Future_q->next; tptr != Future_q; tptr = tptr->next)
	    if (tptr->ufunc == ufunc
                && tptr->time == simtime
                && tptr->pid == pid)
		return tptr;
    }

    /* if we get here, then we did not find a matching function pointer */
    return NULL;
}

/*
 * This routine prints out the entries in the time wheel. It uses the
 * symbol table information to lookup the names of functions corresponding
 * to function addresses.
 */
void
wheel_print ()
{
    task_ptr pnew;
    int pos;
    /*    func_name_ptr fptr; */
    char *fname;

#if defined(DEBUG) || defined(NO_INLINING)
    fprintf (stderr, "Curpos = %d, ", Curptr - &Wheel[0]);
#endif
#ifndef ULONG_TIME
    fprintf (stderr, "End_time = %.1f, Start_time = %.1f\n",
	     End_time, Start_time);
#else
    fprintf (stderr, "End_time = %u\n", End_time);
#endif
    for (pos = 0; pos < WHEEL_SIZE; pos++) {
	pnew = Wheel[pos].next;
	if (pnew != (task_ptr) & Wheel[pos]) {
	    fprintf (stderr, "%d:", pos);
	    while (pnew != (task_ptr) & Wheel[pos]) {
		/*
		 * fptr = (func_name_ptr) hash_lookup(Func_hash_tab,
		 * (ulong) pnew->ufunc);
		 * if (fptr)
		 * fname = fptr->fname;
		 * else */
		fname = "(unknown)";
#ifndef ULONG_TIME
		fprintf (stderr, "\tp%d time %.1f, pri %d, func 0x%x %s,\n",
			 pnew->pid, pnew->time, pnew->priority, pnew->ufunc,
			 fname);

#else
		fprintf (stderr, "\tp%d time %u, pri %d, func 0x%x %s,\n",
			 pnew->pid, pnew->time, pnew->priority, pnew->ufunc,
			 fname);
#endif
		pnew = pnew->next;
	    }
	    fprintf (stderr, "\n");
	}
    }
    pnew = Future_q->next;
    if (pnew == Future_q)
	return;
    fprintf (stderr, "Future_q:\n");
    for (; pnew != Future_q; pnew = pnew->next) {
	/*
	 * fptr = (func_name_ptr) hash_lookup(Func_hash_tab, (ulong)
	 * pnew->ufunc); 
	 * if (fptr)
	 * fname = fptr->fname;
	 * else */
	fname = "(unknown)";
#ifndef ULONG_TIME
	fprintf (stderr, "\tp%d time %.1f, pri %d, func 0x%x %s,\n",
		 pnew->pid, pnew->time, pnew->priority, pnew->ufunc,
		 fname);
#else
	fprintf (stderr, "\tp%d time %u, pri %d, func 0x%x %s,\n",
		 pnew->pid, pnew->time, pnew->priority, pnew->ufunc,
		 fname);
#endif
    }
}
